# coding: utf-8

require './header-footer.rb'

class Numeric
  def to_tex
    if (self.is_a?(Rational))
      return "\\frac{#{self.numerator}}{#{self.denominator}}"
    end
    return to_s
  end
  def simplify(what=nil)
    self
  end
end

class Fixnum
  def simplify(what=nil)
    self
  end
  def to_tex
    return to_s
  end
end

class Term
  
  def self.standard
    {:trivial=>true, :all=>true, :doubleFrac=>true, :prodpowExpand => true, :powpowExpand => true, :simplify => true, :negPowToFrac => true}
  end

  def *(t)
    return Prod.new([self,t])
  end
  def **(t)
    return Power.new(self,t)
  end
  def /(t)
    return Frac.new(self,t)
  end
  def getVars
    if self.is_a?(Var)
      return {self.name => self.power}
    elsif self.is_a?(Prod)
      vars = Hash.new
      self.terms.select{|t| t.is_a?(Var)}.each{|v| vars[v.name] ? vars[v.name]+=v.power : vars[v.name]=v.power}
      return vars
    end
    return nil
  end
end

# simplifications:
#
# :negPowToFrac
# :prodpowExpand
# :powpowExpand
# :doubleFrac
# :simplify



class Var < Term
  attr_reader :name, :power
  def initialize(name, power=1)
    @name = name
    @power = power
  end
  def simplify(what=Term.standard)
    if (what[:all] || what[:trivial]) && (@power==0)
      return 1
    end
    self
  end
  def inv
    return Var.new(@name,-@power)
  end
  
  def to_tex()
    return "#{@name}" + (@power!=1 ? "^{#{@power}}" : "")
  end
  def to_s
    return "Var[#{@name}^#{@power}]"
  end
end

class Prod < Term
  attr_reader :terms
  def initialize(terms)
    @terms = Array.new(terms)
  end
  def simplify(what=Term.standard)
    # puts "Prod simplify terms: #{@terms.inspect}"
    # flatten and simplify first
    puts "     Terms "+@terms.inspect
    @terms = @terms.map{|t| r = t.simplify(what); r.is_a?(Prod) ? r.terms : r }.flatten
    puts "     -> Terms "+@terms.inspect
    # puts "  flattend and simplified: #{@terms.inspect}"
    # collect variables
    vars = getVars
    # collect fractions
    fracs = @terms.select{|t| t.is_a?(Frac)}
    nums = fracs.map{|f| f.num}
    denoms = fracs.map{|f| f.denom}
    # collect numbers
    coefficient = @terms.select{|t| t.is_a?(Numeric)}.inject(1){|a,b| a*b}
    if coefficient==0
      return 0
    end
    # and the rest
    rest = @terms.select{|t| !(t.is_a?(Frac) || t.is_a?(Var))}
    #
    ##puts "Collected items:"
    ##puts "  vars:   #{vars.inspect}"
    ##puts "  nums:   #{nums.inspect}"
    ##puts "  denoms: #{denoms.inspect}"
    ##puts "  coef:   #{coefficient}"
    ##puts "  rest:   #{rest.inspect}"
   
    # Build the new product
    nterms = []
    if coefficient!=1
      nterms.push(coefficient)
    end
    vars.keys.sort.each{|v| nterms.push(Var.new(v, vars[v]))}
    nterms+=nums
    nterms+=rest
    ##puts "  nterms: #{nterms.inspect}"
    if (fracs.size>0)
      return Frac.new(Prod.new(nterms), Prod.new(denoms))
    end
    if nterms.size==1
      return nterms[0]
    end
    res = Prod.new(nterms)
    if (what[:all] || what[:negPowToFrac]) && res.is_a?(Prod)
      return res.negPowToFrac
    end
    return res
  end

  def negPowToFrac
    negPows = @terms.select{|t| (t.is_a?(Power) || t.is_a?(Var)) && t.power<0}.map{|p| p.inv}
    if (negPows.size>0)
      rest = @terms.select{|t| !((t.is_a?(Power) || t.is_a?(Var)) && t.power<0)}
      return Frac.new(Prod.new(rest), Prod.new(negPows));
    else
      return self
    end
  end
  
  def to_tex()
    return @terms.map{|t| t.to_tex}.join(" \\cdot ")
  end
  def to_s
    return "Prod["+@terms.map{|t| t.to_s}.join(", ")+"]"
  end
end

class Power < Term
  attr_reader :base, :power
  def initialize(base, power)
    @base = base
    @power = power
  end
  def inv
    return Power.new(@base,-@power)
  end
  def simplify(what=Term.standard)
    puts "  Base = "+@base.to_s
    @base = @base.simplify
    puts "  -> Base = "+@base.to_s
    if (@power == 0)
      return 1
    end
    if (@power == 1)
      return @base
    end
    if (@base.is_a?(Numeric) && @power.is_a?(Numeric))
      return @base**@power
    end
    if (what[:prodpowExpand] || what[:all])
      if (@base.is_a?(Prod))
        return Prod.new(@base.terms.map{|t|
                          Power.new(t,power)
                        })
      end
      if (@base.is_a?(Frac))
        if (@power>1)
          return Frac.new(Power.new(@base.num, @power), Power.new(@base.denom, @power))
        else
          return Frac.new(Power.new(@base.denom, -@power), Power.new(@base.num, -@power))
        end
      end
    end
    if (what[:powpowExpand] || what[:all])
      if (@base.is_a?(Power))
        return Power.new(@base,Prod.new([@base.power, @power]))
      end
      if (@base.is_a?(Var))
        return Var.new(@base.name, @base.power*@power)
      end
    end
    return self
  end

  def to_tex
    p = (@base.is_a?(Prod) || @base.is_a?(Frac) || @base.is_a?(Power) || (@base.is_a?(Var) && @base.power!=1))
    p1 = p ? "\\left(" : ""
    p2 = p ? "\\right)" : ""
    return "#{p1}#{@base.to_tex}#{p2}^{#{@power.to_tex}}"
  end

  def to_s
    return "Power[#{@base.to_s}, #{@power.to_s}]"
  end
end

class Frac < Term
  attr_reader :num, :denom
  def initialize(num, denom)
    @num = num
    @denom = denom
  end
  def simplify(what=Term.standard)
    @num = @num.simplify(what)
    @denom = @denom.simplify(what)
    if (@denom==1)
      return @num
    end
    if (@denom==-1)
      return Prod.new(-1,@num).simplify(what)
    end
    if (@denom.is_a?(Numeric))
      return Prod.new(Rational(1)/@denom,@num).simplify(what)
    end
    if (what[:doubleFrac] || what[:all])
      if (@denom.is_a?(Frac))
        return Prod.new([@num, Frac.new(@denom.denom, @denom.num)])
      end
    end
    if (what[:simplify] || what[:all])
      res = Prod.new([@num, Power.new(@denom,-1)]).simplify(what)
      if res.is_a?(Prod)
        res =res.negPowToFrac
      end
      return res
    end
    return self
  end

  
  def to_tex
    return "\\frac{#{@num.to_tex}}{#{@denom.to_tex}}"
  end
  def to_s
    return "Frac[#{@num.to_s}, #{@denom.to_s}]"
  end
end


def simpleProd(vars)
  e = vars[0]**(rand(11)-5)
  for i in 1...(vars.size)
    e*=vars[i]**(rand(11)-5)
  end
  return e**(rand(11)-5)
end

def simpleFrac(vars)
  (simpleProd([vars[rand(vars.size)],vars[rand(vars.size)]])/simpleProd([vars[rand(vars.size)],vars[rand(vars.size)]]))**(rand(11)-5)
end

vars = [Var.new("a"),  Var.new("b"), Var.new("c")]




ex = []
sol = []

1.times {
  e = e = simpleProd(vars) # simpleFrac(vars)
  puts e
  c =e.to_tex
  ex.push(c)
  sol.push([c])
  while (true)
    e=e.simplify
    d = e.to_tex
    break if (d==c)
    puts e
    c = d
    sol[-1].push(c)
  end
}

quit
name = "dbruch"

puts "Vereinfachen Sie soweit wie möglich. Schreiben Sie das Resultat als Bruch aus Produkten aus Potenzen mit positiven Exponenten."
puts header(name);
p (ex.map{|e| "$"+e+"$"}).zip(sol.map{|e| "$"+e.join(" = ")+"$"})
puts footer(name);
